"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const json_store_1 = require("./json-store");
const utils_1 = require("./utils");
class Index {
    constructor(filename) {
        this.filename = filename;
        this.maxHeight = 32;
        this.logger = utils_1.logger('index');
        this.store = new json_store_1.default(filename, 0);
    }
    async create() {
        const entry = new IndexEntry("", 0, new SkipListNode([]));
        await this.store.create([entry.serialized()]);
    }
    async open() {
        await this.store.open();
    }
    async close() {
        await this.store.close();
    }
    async drop() {
        await this.store.destroy();
    }
    get isOpen() {
        return this.store.isOpen;
    }
    async lastModified() {
        return await this.store.lastModified();
    }
    async beginTransaction(field) {
        const head = await this.lockHead(field, true);
        const value = JSON.parse(head.node.value);
        if (value.tx)
            throw new IndexError(`Field "${field}" already in transaction`);
        value.tx = 1;
        head.node.value = JSON.stringify(value);
        await this.updateEntry(head);
        await this.unlockHead(head);
    }
    async endTransaction(field) {
        const head = await this.lockHead(field, true);
        const value = JSON.parse(head.node.value);
        if (!value.tx)
            throw new IndexError(`Field "${field}" not in transaction`);
        value.tx = 0;
        head.node.value = JSON.stringify(value);
        await this.updateEntry(head);
        await this.unlockHead(head);
    }
    async insert(objectFields, cache = new Map()) {
        if (!Array.isArray(objectFields))
            objectFields = [objectFields];
        if (!objectFields.length)
            return;
        const fieldName = objectFields[0].name;
        const head = await this.lockHead(fieldName, true, cache);
        const info = JSON.parse(head.node.value);
        const isDateTime = info.type == 'date-time';
        const transform = (o) => {
            if (isDateTime && typeof o.value == 'string')
                o.value = Date.parse(o.value);
        };
        if (objectFields.length == 1) {
            transform(objectFields[0]);
        }
        else {
            // Sort in descending order to allow a single insert
            objectFields.sort((b, a) => {
                transform(a);
                transform(b);
                return a.value < b.value ? -1 : a.value > b.value ? 1 : 0;
            });
        }
        const inserts = [];
        const updates = new Set();
        this.logger.time(`traversing index entries - ${fieldName}`);
        for (const objectField of objectFields) {
            const positions = await this.indexObjectField(objectField, head, cache, inserts);
            for (const position of positions)
                if (position > 0)
                    updates.add(position);
        }
        this.logger.timeEnd(`traversing index entries - ${fieldName}`);
        // Lock writes completely until we're done to ensure the append position
        // remains correct
        await this.store.lock(0, { exclusive: true });
        const { position: startPosition } = await this.store.getAppendPosition();
        let insertPosition = startPosition;
        const pendingRaw = [];
        const offset = this.store.joiner.length;
        const process = (entry) => {
            const position = insertPosition + offset;
            entry.position = position;
            for (let i = 0; i < entry.node.levels.length; i++) {
                const pos = entry.node.levels[i];
                if (pos >= 0)
                    continue;
                const next = inserts[-pos - 1];
                entry.node.levels[i] = next.position;
            }
            const raw = this.store.stringify(entry.serialized());
            insertPosition = position + raw.length;
            pendingRaw.push(raw);
        };
        this.logger.time(`inserting index entries - ${fieldName}`);
        for (let i = 0; i < inserts.length; i++) {
            const entry = inserts[i];
            // Insert all the duplicates first so that the main entry can link
            // to the first duplicate
            const value = entry.node.value;
            let prev = null;
            let dupe = null;
            if (entry.node.isDuplicate) {
                process(entry);
                prev = entry;
            }
            while ((dupe = inserts[i + 1]) && dupe.node.value == value) {
                if (prev)
                    dupe.link = prev.position;
                process(dupe);
                prev = dupe;
                ++i;
            }
            if (!entry.node.isDuplicate) {
                if (prev)
                    entry.link = prev.position;
                process(entry);
            }
        }
        this.logger.timeEnd(`inserting index entries - ${fieldName}`);
        if (pendingRaw.length)
            await this.store.appendRaw(pendingRaw.join(this.store.joiner), startPosition);
        await this.store.unlock();
        // Update the cache after unlocking as it is slow
        for (let i = 0; i < inserts.length; i++) {
            const entry = inserts[i];
            cache.set(entry.position, entry);
        }
        this.logger.time(`updating index entries - ${fieldName}`);
        for (const pos of updates) {
            const entry = cache.get(pos);
            for (let i = 0; i < entry.node.levels.length; i++) {
                const p = entry.node.levels[i];
                if (p < 0)
                    entry.node.levels[i] = inserts[-p - 1].position;
            }
            if (entry.link < 0)
                entry.link = inserts[-entry.link - 1].position;
            await this.updateEntry(entry);
        }
        this.logger.timeEnd(`updating index entries - ${fieldName}`);
        await this.unlockHead(head);
    }
    async indexObjectField(objectField, head, cache = new Map(), inserts) {
        const { name, value, position } = objectField;
        let height = head.node.levels.filter(p => p != 0).length;
        const maxLevel = Math.min(height, this.maxHeight - 1);
        let level = 0;
        while (level < maxLevel && Math.round(Math.random()))
            ++level;
        height = Math.max(height, level + 1);
        const updates = [];
        let current = head;
        for (let i = height - 1; i >= 0; i--) {
            let nextNodePos;
            while (nextNodePos = current.node.next(i)) {
                // Check cache ourselves to avoid promise overhead
                const next = nextNodePos < 0 ? inserts[-nextNodePos - 1] :
                    cache.get(nextNodePos) || await this.getEntry(nextNodePos, cache);
                if (next.node.value <= value)
                    current = next;
                if (next.node.value >= value)
                    break;
            }
            if (i > level)
                continue;
            updates.push(current);
        }
        const prev = updates[updates.length - 1];
        const isDuplicate = prev.node.value == value;
        const entry = isDuplicate ?
            new IndexEntry(name, position, new SkipListNode([], value)) :
            new IndexEntry(name, position, new SkipListNode(Array(level + 1).fill(0), value));
        inserts.push(entry);
        entry.position = -inserts.length; // placeholder position
        if (isDuplicate) {
            entry.link = prev.link;
            prev.link = entry.position;
            return new Set([prev.position]);
        }
        const positions = new Set();
        for (let i = 0; i <= level; i++) {
            const current = updates[updates.length - i - 1];
            entry.node.levels[i] = current.node.next(i);
            current.node.levels[i] = entry.position;
            positions.add(current.position);
        }
        return positions;
    }
    async getFields() {
        let head = await this.getRootEntry();
        const fields = [];
        while (head.link) {
            head = await this.getEntry(head.link);
            const info = JSON.parse(head.node.value);
            fields.push({
                name: head.field,
                ...info
            });
        }
        return fields;
    }
    async addFields(fields) {
        if (!fields.length)
            return;
        let head = await this.getRootEntry();
        while (head.link) {
            head = await this.getEntry(head.link);
            fields = fields.filter(f => f.name != head.field);
        }
        let position;
        for (const field of fields) {
            const prevHead = head;
            const info = { tx: 0 };
            if (field.type == 'date-time')
                info.type = field.type;
            head = new IndexEntry(field.name, 0, new SkipListNode(Array(this.maxHeight).fill(0), JSON.stringify(info)));
            position = await this.insertEntry(head, undefined, position);
            prevHead.link = head.position;
            await this.updateEntry(prevHead);
        }
    }
    async lockHead(field, exclusive = false, cache) {
        let head = await this.getRootEntry(cache);
        while (head.field != field && head.link) {
            head = await this.getEntry(head.link, cache);
            if (head.field == field) {
                // Lock and get the entry again since it might have changed
                // just before locking
                await this.store.lock(head.position, { exclusive });
                head = await this.getEntry(head.position, cache);
            }
        }
        if (head.field != field)
            throw new IndexError(`Field "${field}" missing from index`);
        return head;
    }
    async unlockHead(head) {
        await this.store.unlock(head.position);
    }
    async getRootEntry(cache) {
        const cached = cache && cache.get(0);
        if (cached)
            return cached;
        const { start, value } = await this.store.get(1);
        const entry = new IndexEntry(value);
        entry.position = start;
        if (cache)
            cache.set(0, entry);
        return entry;
    }
    async insertEntry(entry, cache, position) {
        const { start, length } = await this.store.append(entry.serialized(), position);
        entry.position = start;
        if (cache)
            cache.set(entry.position, entry);
        return start + length;
    }
    async getEntry(position, cache) {
        const cached = cache && cache.get(position);
        if (cached)
            return cached;
        const { start, value } = await this.store.get(position);
        const entry = new IndexEntry(value);
        entry.position = start;
        if (cache)
            cache.set(entry.position, entry);
        return entry;
    }
    async updateEntry(entry) {
        // offset 4 = 1 brace + 2 quotes + 1 colon
        await this.store.set(entry.position + 4 + entry.field.length, entry.encoded());
    }
    async find(field, predicate) {
        const cache = new Map();
        const head = await this.lockHead(field, false, cache);
        const height = head.node.levels.filter(p => p != 0).length;
        const info = JSON.parse(head.node.value);
        if (info.tx)
            throw new IndexError(`Field "${field}" in transaction`);
        if (info.type == 'date-time')
            predicate.key = Date.parse;
        let found = false;
        let current = head;
        for (let i = height - 1; i >= 0; i--) {
            let nextNodePos;
            while (nextNodePos = current.node.next(i)) {
                const next = await this.getEntry(nextNodePos, cache);
                const { seek } = predicate(next.node.value);
                if (seek <= 0)
                    current = next;
                if (seek == 0)
                    found = true;
                if (seek >= 0)
                    break;
            }
            if (found)
                break;
        }
        if (current == head)
            current = current.node.next(0) ?
                await this.getEntry(current.node.next(0), cache) : null;
        const pointers = new Set();
        while (current) {
            let entry = current;
            current = current.node.next(0) ?
                await this.getEntry(current.node.next(0), cache) : null;
            const { seek, match } = predicate(entry.node.value);
            if (seek <= 0 && !match)
                continue;
            if (!match)
                break;
            pointers.add(entry.pointer);
            while (entry.link) {
                const link = await this.getEntry(entry.link, cache);
                pointers.add(link.pointer);
                entry = link;
            }
        }
        await this.unlockHead(head);
        return pointers;
    }
}
class IndexError extends Error {
}
exports.IndexError = IndexError;
IndexError.prototype.name = 'IndexError';
class SkipListNode {
    get type() {
        return typeof this.value == 'boolean' ? 1 /* Boolean */ :
            typeof this.value == 'number' ? 2 /* Number */ :
                typeof this.value == 'string' ? 3 /* String */ :
                    0 /* Null */;
    }
    constructor(obj, value) {
        if (Array.isArray(obj)) {
            const levels = obj;
            if (!levels)
                throw new TypeError('levels is required');
            if (typeof value == 'number' && !Number.isFinite(value))
                throw new TypeError('Number value must be finite');
            this.value = value == null ? null : value;
            this.levels = levels;
        }
        else {
            const encodedNode = obj;
            const parts = encodedNode.split(';');
            const [encodedLevels, encodedType] = parts.slice(0, 2);
            const type = utils_1.z85DecodeAsUInt32(encodedType, true);
            this.value = SkipListNode.decodeValue(type, parts.slice(2).join(';'));
            this.levels = encodedLevels ?
                encodedLevels.split(',').map(l => utils_1.z85DecodeAsUInt32(l)) : [];
        }
    }
    get isDuplicate() {
        return this.levels.length == 0;
    }
    next(level) {
        return this.levels[level];
    }
    encoded() {
        if (this.isDuplicate)
            return ';;';
        const encodedLevels = this.levels.map(l => utils_1.z85EncodeAsUInt32(l)).join(',');
        const encodedType = utils_1.z85EncodeAsUInt32(this.type, true);
        let encodedValue = '';
        if (typeof this.value == 'boolean')
            encodedValue = utils_1.z85EncodeAsUInt32(Number(this.value), true);
        else if (typeof this.value == 'number')
            encodedValue = utils_1.z85EncodeAsDouble(this.value, true);
        else if (typeof this.value == 'string')
            encodedValue = this.value;
        return `${encodedLevels};${encodedType};${encodedValue}`;
    }
    static decodeValue(type, value) {
        switch (type) {
            case 1 /* Boolean */:
                return Boolean(utils_1.z85DecodeAsUInt32(value, true));
            case 2 /* Number */:
                return utils_1.z85DecodeAsDouble(value, true);
            case 3 /* String */:
                return value;
            default:
                return null;
        }
    }
}
class IndexEntry {
    constructor(obj, pointer, node) {
        this.position = 0;
        this.link = 0; // Pointer to next duplicate in index
        if (typeof obj == 'string') {
            const field = obj;
            if (pointer == null || !node)
                throw new TypeError('pointer and node are required');
            this.field = field;
            this.pointer = pointer;
            this.node = node;
        }
        else {
            const serializedEntry = obj;
            this.field = Object.keys(serializedEntry)[0];
            const encodedParts = serializedEntry[this.field].split(';');
            const encodedPointer = encodedParts[0];
            const encodedLink = encodedParts[1];
            const encodedNode = encodedParts.slice(2).join(';');
            this.pointer = utils_1.z85DecodeAsUInt32(encodedPointer);
            this.link = utils_1.z85DecodeAsUInt32(encodedLink);
            this.node = new SkipListNode(encodedNode);
        }
    }
    encoded() {
        const encodedPointer = utils_1.z85EncodeAsUInt32(this.pointer);
        const encodedLink = utils_1.z85EncodeAsUInt32(this.link);
        return `${encodedPointer};${encodedLink};${this.node.encoded()}`;
    }
    serialized() {
        return {
            [this.field]: this.encoded()
        };
    }
}
exports.default = Index;
//# sourceMappingURL=index.js.map